#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <unistd.h>


#include <windows.h>
#include <winnt.h>

#include "doa.h"


struct doa_func doa104p2 = {
    .randomize_char_list = (void*)0,
    .get_num_faces = (void*)0x1151A0,
    .get_num_alts = (void*)0x115270,
    .random = (void*)0x3D7110,
    .PlayerCfgTbl = (void*)0xEC0120,
    .get_plcfg_char = (void*)0x654990,
    .get_plcfg_cos = (void*)0x6549F0,
    .get_something = (void*)0x3D4E50,
    .patch1 = 0x6905fe,
    .patch2 = 0x3e1eb9,
    .patch3 = 0x2f88b3,
};

struct doa_func doa104p3 = {
    .randomize_char_list = (void*)0,
    .get_num_faces = (void*)0x114de0,
    .get_num_alts = (void*)0x114eb0,
    .random = (void*)0x3d7960,
    .PlayerCfgTbl = (void*)0xEC0140,
    .get_plcfg_char = (void*)0x655200,
    .get_plcfg_cos = (void*)0x655260,
    .get_something = (void*)0x3d56b0,
    .patch1 = 0x69139E,
    .patch2 = 0x3e2459,
    .patch3 = 0x2f9483,

};

struct doa_func doa104p5 = {
    .randomize_char_list = (void*)0,
    .get_num_faces = (void*)0x114f30,
    .get_num_alts = (void*)0x115000,
    .random = (void*)0x3d7780,
    .PlayerCfgTbl = (void*)0xEC0140,
    .get_plcfg_char = (void*)0x655200,
    .get_plcfg_cos = (void*)0x655260,
    .get_something = (void*)0x3d54d0,
    .patch1 = 0x69139e,
    .patch2 = 0x3e2439,
    .patch3 = 0x2f95f3,
};

static struct supported {
    const char *ver;
    struct doa_func *fn_tbl;
} supported[] = { {"1.04p2", &doa104p2}, {"1.04p3",&doa104p3},{"1.04p4",&doa104p3},{"1.04p5", &doa104p5}};

static struct info *g_Nfo;
struct PlayerCfg {
    BYTE ch_id, cs_id, head_id, glasses, alt_id;
} __attribute__((packed));

static void randomize_costume(struct PlayerCfg *cfg, int do_alt)
{
    struct cos_entry *cos = get_cos(cfg->ch_id, cfg->cs_id);
    struct char_csopts_entry *opts = &DOA.data.opts_tbl[cfg->ch_id];

    uint8_t nf,gl,nalt;

    /* get info what's available for this costume */
    nf = DOA.func.get_num_faces(cfg->ch_id, cfg->cs_id);
    gl = opts->glasses[cfg->cs_id];
    nalt = DOA.func.get_num_alts(cfg->ch_id, cfg->cs_id);

    if(nf>1) cfg->head_id=DOA.func.random(0, nf-1);
    // make no glasses a bit more likely
    if(gl==2) cfg->glasses= (DOA.func.random(0, 2) == 0);

    /* can't rely on get_num_alts, it is 1 even if there is are details */
    if(do_alt && cos->alt_tbl) {
            if(nalt > 1) cfg->alt_id=DOA.func.random(0, nalt-1);
    }
    pr_dbg("ch:%d cs:%d face:%d/%d alt:%d/%d gl:%d/%d\n",cfg->ch_id,cfg->cs_id,
           cfg->head_id+1,nf, (cos->alt_tbl)?cfg->alt_id+1:0,
           (cos->alt_tbl)?nalt:0, cfg->glasses, (gl==2)?1:0);
}

static int __stdcall fixup_cpu_costumes(DWORD *dst, DWORD *src)
{
    pr_dbg("*patching random cpu opponents(%d)\n",dst[-1]);
    int res = DOA.func.randomize_char_list(dst, src);

    struct PlayerCfg *cfg = (void*)(dst[0]);
    struct PlayerCfg *end = (void*)(dst[1]);

    for( ; cfg < end; cfg++)
        randomize_costume(cfg, TRUE);

    return res;
}

static int __stdcall fixup_player_costumes(void)
{
    pr_dbg("*randomize player costumes\n");
    struct PlayerCfg *cfg = (void*)DOA.func.PlayerCfgTbl;

    for(int team = 0; team < 2; team++) {
        for(int idx = 0; idx<0xf; idx++) {
            /* note: this is the setting the player chose in the selection
             * screen */
            int sel_ch = DOA.func.get_plcfg_char(team, idx);
            if( sel_ch == 0xff) continue;
            int sel_cs = DOA.func.get_plcfg_cos(team, idx);

            /* sometimes in versus mode the game hasn't picked a costume/char
             * yet?? can't do anything then */
            if( (cfg->ch_id == 0xff) || (cfg->cs_id==0xff) ) continue;

            /* if the character or the costume was chosen randomly */
            if( (sel_ch == 0xFD) || (sel_cs == 0xFE) ) {
                pr_dbg("%d/%d ",team, idx);
                /* note: game chooses random underwear texture here */
                randomize_costume(cfg, FALSE);
                        }
            cfg++;
        }
    }
    return DOA.func.get_something(2);
}


static void* swap_rel_target(DWORD *dst, void *target)
{

    DWORD old_target = dst[0]+(DWORD)dst+4;
    DWORD rel_ofs = (DWORD)target - (DWORD)dst - 4;
    dst[0]=rel_ofs;
    return (void*)old_target;
}


struct patch_info {
    const char *desc;
    DWORD marker_len;
    const BYTE *marker;
    int (*apply)(void* mem);
};

/* show selected Details at character selection screen
 * patch jz -> jnz */
static int apply_patch1(void *mem)
{
    *((BYTE *)mem+2) = 0x75;
    return 0;
}



static const BYTE marker[] = {0x84, 0xC0, 0x74, 0x3, 0x8A, 0x5F, 0xf, 0x8a };
static const struct patch_info patch1 = {
    "show selected Details on character selection screen",
    sizeof(marker), marker, apply_patch1 };


/* random hairstyle for cpu oppenents in arcade/time attack/survival
 * replace: call <...> with our own func */
static int apply_patch2(void *mem)
{
    void *f = swap_rel_target((DWORD*)mem + 1, (void*)&fixup_cpu_costumes);
    DOA.func.randomize_char_list = f;
    return 0;
}
static const BYTE marker2[] = { 0x50, 0x8B,0xCB, 0xE8 }; //0x9F,0xDF, 0xFF, 0xFF };
static const struct patch_info patch2 = {
    "select random head/glasses/details for cpu player",
    sizeof(marker2),marker2, apply_patch2};


/* random hair/... for player */
static const BYTE marker3[] =
{ 0x80,0x7D,0xE4,0x00,  /* cmp byte [ebp-1C],0*/
  0x74,0x34,          /* 0,6 jz .loc_2F88ED*/
  0x6A,0x2,           /* 6,2 push 2 */
  0xC6,0x5};  //,0x2F,0xAA,0x37,0x1,0x1,   /*8,7 mov [0x137aa2f],1*/
 // 0xE8,0x89,0xC5,0xD,0x0};  /*15,5call 0x3D4E50 -> 20bytes */
 /* note: don't match remove last part, absolute address will get fixed by
  * windows to match the load address of game.exe */

static const BYTE patch3_code[] =
{ 0xE8,0xff,0xff,0xff,0xff, /* .0  call my_func        (5) !! fix addr */
  0x80,0x7D,0xE4,0x00,      /* .5  cmp byte [ebp-1C],0 (4) */
  0x74,0x2F, /*0x34-5=0x2F*//* .9 jz .loc_2F88ED       (2) */
  0x90,0x90,                /* .B                      (2) */
  0xC6,0x5,0xff,0xff,0xff,0xff,0x1}; /* .D mov [0x137aa2f],1 (7) !! fix addr*/

 /* do the call in our own func and return that eax
 * branch of jz jump ignores eax so that's no problem */

static int apply_patch3(void *p)
{
    BYTE *mem = p;
    DWORD abs_ofs = *((DWORD*)(mem+10));
    memcpy(mem, patch3_code, sizeof(patch3_code));

    swap_rel_target( (void*)(mem+1), (void*)&fixup_player_costumes);
    *((DWORD*)(mem+0xF)) = abs_ofs;
    return 0;
}
static const struct patch_info patch3 = {
    "select random head/glasses/details for human player",
    sizeof(marker3), marker3, apply_patch3
};

static int load_patch(const struct patch_info *patch, DWORD addr)
{
    int ret;
    BYTE *mem = (BYTE*)addr;
    pr_log("applying patch: %s\n", patch->desc);
    if( memcmp(patch->marker, mem, patch->marker_len) ) {
        pr_log("*error* code is different then expected\n",mem);
        return -1;
    }
    ret = patch->apply(mem);
    if( ret )
        pr_log("*error* failed to apply\n");
    return ret;
}


int doa_plus_init(struct info *nfo)
{
    g_Nfo = nfo;
    DWORD *f_ptr = NULL;

    pr_log("doa++ init ...\n");

    for(int i = 0; i < ARRAY_SIZE(supported); i++) {
        if( strcmp(g_Nfo->exe->name, supported[i].ver) == 0) {
            f_ptr = (DWORD*)supported[i].fn_tbl;
            break;
        }
    }
    if( f_ptr == NULL ) {
        pr_log("doa++  disabled. game version not supported\n");
        return 0;
    }

    /* fixup doa_func stuff to game.exe's memory location */
    DWORD *p = (DWORD*)(&DOA.func);
    for(int i = 0; i<sizeof(DOA.func); i+=sizeof(DWORD),f_ptr++,p++)
        *p = (DWORD)DOA.base + *f_ptr;

    load_patch( &patch1, DOA.func.patch1 ); // fix character selection
    load_patch( &patch2, DOA.func.patch2 ); // give cpu player random hair/glasses
    load_patch( &patch3, DOA.func.patch3 ); // give player random hair/glasses/details

    return 0;
}
