#ifndef DOA_H
#define DOA_H



#define ARRAY_SIZE(a) (sizeof(a)/sizeof((a)[0]))
#define VERSION "0.1.4"


/*************************
 * DOA5 internal structure definitions
 */


struct asset_entry {
    const char *name;
    uint32_t    type;
    const char *lnk;
    uint32_t    update_idx; /*LOWORD is idx of the asset replaced or 0xffff
			      if it isn't an update. HIWORD is always 0xffff*/
}__attribute__((packed));

#define FLAG_XOR (1<<31)
#define FLAG_ZIP (1<<30)
#define FLAG_CRC (1<<29)
#define CRC_TBL_IDX(flags) ( flags & 0xffff )
struct flag_entry {
    uint32_t size; //contains on-disk size, eventually compressed
    uint32_t flags;
}__attribute__((packed));

struct crc_entry {
    uint32_t size; //unzipped size
    uint32_t crc;
}__attribute__((packed));

/* note: the uint32_t used here to refer to game assets, are indices
 *       into the asset table
 *       unused entries are set to: 0x0000ffff */
struct doa_mdl {
    uint32_t tmc;
    uint32_t tmcl;
}__attribute__((packed));

struct face_entry {
    struct doa_mdl hair;
    struct doa_mdl face;
    struct doa_mdl body;
}__attribute__((packed));

struct alt_entry {
    uint32_t h;
    uint32_t hl;
}__attribute__((packed));

struct cos_entry {
    uint32_t tmc;
    uint32_t tmcl;
    uint32_t c;
    uint32_t p;
    uint32_t phyd;
    struct face_entry *face_tbl; /* number of entries here is listed in the  */
    struct alt_entry *alt_tbl;   /* char_csopts_entry under num_face/num_alt */
}__attribute__((packed));

struct char_entry {
    void *unk_tbl; /* seems to list .bin/mpm/cam/sktm/mot files
		      unclear how to know the number of entries here */
    struct cos_entry *cos_tbl; /* number of entries given by the corresponding
				  char_csopts_entry->num for that character */
    struct doa_mdl fx;
    struct doa_mdl ai;
    uint32_t face_mpm;
    uint32_t phyd;
    uint32_t unk; //??
}__attribute__((packed));

struct slot_entry {
    uint32_t cos_idx;
    uint32_t type; /* probably: 0->base,unlocked, 1->base,locked, 2->dlc */
}__attribute__((packed));

struct char_slot_entry {
    uint32_t num;   // total number of costume slots
    uint32_t num_base; // number of base costumes without DLC_/DLCU_*
    struct slot_entry *slots;
}__attribute__((packed));

struct char_csopts_entry {
    uint8_t num;
    uint8_t num_base;
    uint16_t reserved;
    uint8_t *type; /* probably just 0->base costume, 1->dlc costume */
    uint8_t *num_faces;
    uint8_t *num_alts;
    uint8_t *glasses; /* indicates whether glasses are enabled */
}__attribute__((packed));

/***********************************************************/

struct doa_data {
    struct asset_entry *asset_tbl;
    struct crc_entry *crc_tbl;
    struct flag_entry *flag_tbl;
    struct char_entry *char_tbl;
    struct char_slot_entry *slot_tbl;
    struct char_csopts_entry *opts_tbl;
    struct doa_mdl **hair_tbl;
    struct doa_mdl **face_tbl;
} __attribute__((packed)); /* make us neatly packed so we can loop through it */

struct doa_func {
    int __stdcall (*randomize_char_list)(DWORD *dst, DWORD *src);
    int __stdcall (*get_num_faces)(unsigned int char_idx, unsigned int cos_idx);
    int __stdcall (*get_num_alts)(unsigned int char_idx, unsigned int cos_idx);
    int __cdecl (*random)(int min, int max);
    char *PlayerCfgTbl;
    int __stdcall (*get_plcfg_char)(int team, int idx);
    int __stdcall (*get_plcfg_cos)(int team, int idx);
    int __stdcall (*get_something)(int n);
    DWORD patch1,patch2,patch3;
} __attribute__((packed));


struct doa {
    DWORD base;
    struct doa_data data;
    struct doa_func func;
} __attribute__((packed));
extern struct doa DOA;
extern const unsigned int MAX_CHARS;

struct exe_characteristics {
    const char *name;
    DWORD compat;
    const char *hash;
    DWORD *ofs_table;
    DWORD compat_ver;
};

struct info {
    HMODULE hmod;
    IMAGE_SECTION_HEADER *rdata,*code;
    const struct exe_characteristics *exe;
    uint32_t fmt_ver;
    int char_idx; //current character /costume idx during mod.config parsing
    int cos_idx;
};

//init.c
int check_compat(struct info *info, const char* ver);
void do_print(int lvl, char* fmt, ...);
#define pr_log(fmt...) do_print(1, fmt)
#define pr_dbg(fmt...) do_print(0, fmt)



// in mod.c
int load_config(struct info *info, const char *path);
struct cos_entry* get_cos(unsigned int ch_id, unsigned int cs_id);

// in doa++.c
int doa_plus_init(struct info *nfo);

#endif
