#include <sys/types.h>
#include <sys/stat.h>

#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <unistd.h>

#include <windows.h>
#include <wincrypt.h>
#include <winnt.h>

#include <dinput.h>

#include "doa.h"


static struct info *g_Nfo;
struct doa DOA;

/* helper */
int g_DEBUG=0;
FILE *f_log;
void do_print(int lvl, char* fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);

    if( (lvl==0) && (g_DEBUG==0) ) return;
    vfprintf(f_log, fmt, ap);
    //vfprintf(stdout, fmt, ap);
    fflush(f_log);
    va_end(ap);
}


/***********************************************
 *
 *table offsets inside DOA5's game.exe
 * FIXME: might want to replace that with some pattern&scanning method
 *        all those number get rather bothersome
 *
 *                         ASSET_TBL  CRC_TBL  FLAG_TBL   CHAR_TBL  SLOT     OPTS     HAIR */
static DWORD ofs_v102[] = {0x0bebee0,0x0c3d050,0x0c4f868,0x09ea9c0,0x8592c8,0x859990,0xAC8000};
static DWORD ofs_v102a[]= {0x0bebee0,0x0c3d0b0,0x0c4f8f8,0x09eb790,0x858338,0x859a00,0xAC8000};
static DWORD ofs_v104[] = {0x0d0BC00,0x0d5d680,0x0d6fec8,0x0af1578,0x95a2b8,0x95b980,0xBE7008};
static DWORD ofs_v104p1[]={0x0d0BC00,0x0d5d680,0x0d6fec8,0x0af1328,0x95a338,0x95ba10,0xBE7008};
static DWORD ofs_v104p2[]={0x0d0BC00,0x0d5d680,0x0d6fec8,0x0af1370,0x95a318,0x95b9f0,0xBE7008};
static DWORD ofs_v104p3[]={0x0d0BC00,0x0d5d680,0x0d6fec8,0x0af1128,0x95a348,0x95ba20,0xBE7008};
static DWORD ofs_v104p4[]={0x0d0BC00,0x0d5d680,0x0d6fec8,0x0af1100,0x95a348,0x95ba20,0xBE7008};
static DWORD ofs_v104p5[]={0x0d0BC00,0x0d5d680,0x0d6fec8,0x0af1080,0x95a318,0x95b9f0,0xBE7008};

#define ALIGN(ofs,b) ( ( ((uint32_t)ofs) + (b-1))& ~(b-1) )
#define FACE_TBL(HAIR_TBL) ((void*) ALIGN(HAIR_TBL + MAX_CHARS*4,8) )


static const struct exe_characteristics known_exe[] = {
     {"1.02",  0, "8D363ECCC41F965A37DDC7C253F03BBD", ofs_v102  },
    {"1.02A",  1, "E9CDC93610C8911B764C82968D5EDA74", ofs_v102a },
 {"1.03beta", -1, "41064265427FD1422743698B09B433CA", NULL },
 {"1.03beta2",-1, "9BD7E0D1E4E1D2E4AB7FE93792F3EF5C", NULL },
     {"1.04",  3, "8DB41D9F65D9C6BDD34AF92FCA60A399", ofs_v104  },
   {"1.04p1",  3, "1F3CB895E42DB528CE6FA013E0FD6A28", ofs_v104p1},
   {"1.04p2",  3, "8D66E321C120F3E6FFF45D8194CC7368", ofs_v104p2},
   {"1.04p3",  3, "8015689A664DDA75F78432ADD6071669", ofs_v104p3},
   {"1.04p4",  3, "9B5C53E3858FB9D647C1F9BFFC81E8D7", ofs_v104p4},
   {"1.04p5",  3, "131C4EABEC19883F1A6EBC500670F4E6", ofs_v104p5},
};

#define MD5LEN 16 /* 128bit */
#define MD5STRINGLEN  32
static int calc_hash(const char* buf, DWORD buf_len, char *md5string)
{
    HCRYPTPROV prov = 0;
    HCRYPTHASH hash = 0;
    DWORD nb;
    int ret = -1;
    BYTE hash_val[MD5LEN];

    if( !CryptAcquireContext(&prov, NULL, NULL, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT) ) {
        pr_log("CryptAquireContext failed: %d\n", GetLastError());
        return -1;
    }
    if( !CryptCreateHash(prov, CALG_MD5, 0, 0, &hash) ) {
        pr_log("CryptCreateHash failed: %d", GetLastError());
        CryptReleaseContext(prov, 0);
        return -1;
    }
    if( !CryptHashData( hash, (const BYTE*)buf, buf_len, 0) ) {
        pr_log("CryptHashData failed: %d\n", GetLastError());
        goto out;
    }

    nb = MD5LEN;
    if( CryptGetHashParam(hash, HP_HASHVAL, hash_val, &nb, 0) ) {
        for(int i = 0; i < nb; i++, md5string+=2)
            snprintf(md5string, 2, "%02X", hash_val[i] );
        ret=0;
    } else {
        pr_log("CryptGetHashParam failed: %d\n", GetLastError());
    }

out:
    CryptDestroyHash(hash);
    CryptReleaseContext(prov, 0);
    return ret;
}

static char* read_pe_signature(const char* exe_path, DWORD offset, DWORD size)
{
    FILE *f = fopen(exe_path, "rb");
    char *buf = NULL;

    if( f == NULL) {
        pr_log("opening game.exe to read pe signature failed\n");
        fclose(f);
        return NULL;
    }
    if( fseek(f, offset, SEEK_SET) ) {
        pr_log("reading game.exe signature failed\n");
        fclose(f);
        return NULL;
    }

    buf = malloc(size);
    fread(buf, 1, size, f);
    fclose(f);

    return buf;
}

static int check_exe(struct info *info, char* exe_path)
{
    IMAGE_DOS_HEADER *dos = (void*)info->hmod;
    IMAGE_NT_HEADERS *nt = (void*)((char*)info->hmod + dos->e_lfanew);
    IMAGE_DATA_DIRECTORY *secDir = &nt->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_SECURITY];
    char *sig;
    char hash[32];
    int ret = -1;

    if(nt->Signature != IMAGE_NT_SIGNATURE) {
        pr_log("nope - no PE file\n");
        return -1;
    }


    if(secDir->VirtualAddress == 0) {
        pr_log("failed to recognise the game.exe\n");
        return -1;
    }

    sig = read_pe_signature(exe_path, secDir->VirtualAddress, secDir->Size);
    if(sig == NULL)
        return -1;

    if( calc_hash( sig, secDir->Size, hash) ) {
        pr_log("calculating game.exe hash failed\n");
        return -1;
    }
    pr_log("exe chksum: %.32s\n",hash);

    /* keep ptr to pe sections for later */
    IMAGE_SECTION_HEADER *sec = IMAGE_FIRST_SECTION(nt);
    for(unsigned int i=0; i<nt->FileHeader.NumberOfSections; i++) {
        if(strncmp((const char*)sec[i].Name,".rdata",IMAGE_SIZEOF_SHORT_NAME) == 0)
            info->rdata=&sec[i];
        else if(strncmp((const char*)sec[i].Name, ".text",IMAGE_SIZEOF_SHORT_NAME) == 0)
            info->code=&sec[i];

    }

    for(unsigned int i=0; i<ARRAY_SIZE(known_exe); i++) {
        if( strncmp(known_exe[i].hash, hash, 32) != 0) continue;

        // found known exe
        info->exe = known_exe + i;
        pr_log("exe version: %s\n", info->exe->name);

        if(info->exe->compat == -1) {
            MessageBox(NULL, "Your version of the game is no longer supported. Please update. (mod loader disabled)",
                       "DoA++ Init - Error", MB_OK);
            pr_log("*warning*: outdated version. abort\n");
            goto out;
        }


        DOA.base = (DWORD)info->hmod;
        DWORD *ptr = (DWORD*) (&DOA.data);
        for(int i = 0; i < ARRAY_SIZE(ofs_v104p2); i++,ptr++)
            *ptr = DOA.base + info->exe->ofs_table[i];
        DOA.data.face_tbl = FACE_TBL(DOA.data.hair_tbl);
        ret = 0;
        goto out;
    }

    pr_log("*error*: unknown exe\n");
    MessageBox(NULL, "Failed to recognize game.exe file.(mod loader disabled)","DoA++ Init - Error", MB_OK);
out:
    if( sig ) free(sig);
    return ret;
}

int check_compat(struct info *info, const char* ver)
{
    pr_log("mod.config for version: %s (exe:%s)\n",ver,info->exe->name);
    if(strcmp(info->exe->name, ver)==0) return 0;

    /* see if we know about the version given in mod.config */
    const struct exe_characteristics *cmp = known_exe;
    for(unsigned int i=0; i<ARRAY_SIZE(known_exe); i++,cmp++) {
        if( strcmp(cmp->name, ver) == 0) {
            if( cmp->compat == info->exe->compat )
                return 0; /* compatible */
            pr_log("*error* version is incompatible with used game version\n");
            return -1;
        }
    }
    pr_log("*error* unknown version\n");
    return -1;
}


static BOOL protect_section(struct info *info, IMAGE_SECTION_HEADER *sec, DWORD prot, DWORD *old_prot)
{
    BOOL res = FALSE;
    DWORD size = sec->Misc.VirtualSize;
    DWORD dummy;
    if( old_prot == NULL) old_prot = &dummy;

    res = VirtualProtect((void*)(sec->VirtualAddress+(char*)info->hmod),
                             size,prot, old_prot);
    if(res == FALSE)
        pr_log("*error* failed to make %s writable: %x\n",sec->Name, GetLastError());
    return res;
}

/********************************************
 * hook real dinput
 */
typedef HRESULT WINAPI (*DI_Create_t)(HINSTANCE, DWORD, REFIID,LPVOID,
                           LPUNKNOWN);
static DI_Create_t Direct8InputCreate_orig;

static int hook_real_dinput(void)
{
    TCHAR path[MAX_PATH];
    HANDLE hmod;

    GetSystemDirectory(path, MAX_PATH);
    lstrcat(path, "\\dinput8.dll");
    hmod = LoadLibrary(path);
    if( hmod == NULL) {
       pr_log("*error* LoadLibrary(dinput8.dll) failed:%x\n",GetLastError());
        return -1;
    }
    pr_log("real dinput8 loaded: %p\n",hmod);
    Direct8InputCreate_orig = (void*)GetProcAddress(hmod,
                            "DirectInput8Create");
    if(Direct8InputCreate_orig==NULL) {
        pr_log("*error* GetProcAddress failed\n");
        return -1;
    }
    return 0;
}

/* do everything here that requires the game.exe to be fully decrypted */
static int runtime_init(void)
{
    DWORD old;
    BOOL res;
    int ret;

    res = protect_section(g_Nfo, g_Nfo->code, PAGE_EXECUTE_READWRITE, &old);

    if(res == FALSE)
        return -1;

    ret = doa_plus_init(g_Nfo);

    protect_section(g_Nfo, g_Nfo->code, old, NULL);

    return ret;
}

#define EXPORT __declspec(dllexport)
typedef HRESULT WINAPI (*DirectInput8Create_t)(HINSTANCE, DWORD, REFIID,LPVOID,
                           LPUNKNOWN);
EXPORT HRESULT WINAPI DirectInput8Create(HINSTANCE hinst, DWORD version,
                     REFIID riid, LPVOID *out, LPUNKNOWN outer)
{

    if(g_Nfo) runtime_init(); //ignore failure

    return Direct8InputCreate_orig(hinst, version, riid, out, outer);
}

static int loadtime_init(void)
{
    TCHAR path[MAX_PATH];
    DWORD n;
    char *exe_path;
    struct info *info = malloc(sizeof(*info));
    int ret = -1;

    memset(info, 0, sizeof(*info));

    n = GetModuleFileName(NULL, path, MAX_PATH);
    if(n==0)
        return -1;
    exe_path = strdup(path); /* keep one copy to read the signature later */

    while(n--) {
        if(path[n]=='\\')
            break;
    }
    path[n+1]='\0'; // keep path delimiter

    /* open log file first */
    lstrcat(path, "mod.log");
    f_log = fopen(path,"w");

    path[n+1]='\0'; // game path
    pr_log("dll version: %s\n", VERSION);
    pr_log("game dir: %s\n", path);

    if( hook_real_dinput() ) goto out;

    info->hmod = GetModuleHandle(NULL);
    pr_log("hmod: %p\n", info->hmod);


    if( check_exe(info, exe_path) ) {
        free(info);
        ret = 0; // just disable us, but continue running the game
        goto out;
    }

    DWORD prot=0;
    if( protect_section(info, info->rdata, PAGE_READWRITE, &prot) == FALSE)
        goto out;

    lstrcat(path, "mod.config");
    if( load_config(info, path) ) {
        MessageBox(NULL, "Failed to load mod.config","Error", MB_OK);
        pr_log("*error* while applying mod.config*\n");
    }

    protect_section(info, info->rdata, prot, NULL);

    g_Nfo = info;
    ret = 0;
out:
    free(exe_path);
    return ret;
}

BOOL WINAPI DllMain(HINSTANCE hinstance, DWORD reason, LPVOID reserved)
{
    if(reason==DLL_PROCESS_ATTACH) {
        DisableThreadLibraryCalls(hinstance);
        if(loadtime_init() == -1) {
            MessageBox(NULL, "Failed to initialize mod loader","Error", MB_OK);
            return FALSE;
        }
    }
    return TRUE;
}

