#include <sys/types.h>
#include <sys/stat.h>

#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <unistd.h>


#include <windows.h>
#include <winnt.h>

#include <dinput.h>

#include "doa.h"


/* keep characters names instead of just an idx number, makes
 * mod.config easier readable */
const char* DOA_CHAR_TBL[]={ "ZACK","TINA","JANNLEE", "EIN", "HAYABUSA",
                             "KASUMI", "GENFU", "HELENA","LEON","BASS",
			     "KOKORO", "HAYATE", "LEIFANG","AYANE","ELIOT",
	                     "LISA","ALPHA152", NULL, NULL, "BRAD",
			     "CHRISTIE","HITOMI", "RTM", NULL,
	                     "BAYMAN", NULL, NULL, NULL, NULL, "RIG",
			     "MILA","AKIRA","SARAH","PAI",NULL, "AYANE_BOSS",
	                     "HAYATE_BOSS", "KASUMI_BOSS", "HAYABUSA_BOSS",
                             "MOMIJI", "RACHEL", "JACKY", "MARIE","PHASE4",
                             "NYOTENGU","HONOKA","RAIDOU" };
const unsigned int MAX_CHARS=ARRAY_SIZE(DOA_CHAR_TBL);

static const char* NOLINK="NO_LINK";


static void* malloc_copy(void *src, unsigned int el_size,
             unsigned int old_num, unsigned int new_num)
{
    void *dst = malloc( el_size * new_num );
    if( src ) { // src might be NULL if the previous list size was zero
        memcpy(dst, src, el_size*old_num);
    }
    /* set new area to zero */
    memset(dst+(el_size*old_num), 0, (new_num-old_num)*el_size);
    return dst;
}

/*
static int set_char_mod(struct info *info, uint32_t char_idx,
            uint32_t *char_def)
{
    struct char_entry *chara = info->char_tbl + char_idx;
    memcpy(&chara->fx, char_def, 7*sizeof(uint32_t));
    return 0;
}
*/

struct cos_entry* get_cos(unsigned int ch_id, unsigned int cs_id)
{
    struct char_entry *ch = DOA.data.char_tbl + ch_id;
    return ch->cos_tbl + cs_id;
}

static int enable_glasses(struct info *info, uint32_t enable)
{
    struct char_csopts_entry *opts = DOA.data.opts_tbl + info->char_idx;

    pr_dbg(" .glasses: %d\n",enable);
    opts->glasses[info->cos_idx] = (enable == 1) ? 2: 1;
    return 0;
}


static int set_cos_def(struct info *info, uint32_t *cos_def)
{
    pr_dbg(" .set cos_def(%d)...\n",info->cos_idx);
    struct cos_entry *cos = get_cos(info->char_idx, info->cos_idx);
    memcpy(cos, cos_def, 5*sizeof(uint32_t));
    return 0;
}

static int set_dlc_list_len(struct info *info, struct doa_mdl **dlc_lists,
                uint32_t from, uint32_t to)
{
    pr_dbg(".num_dlc_list(%c): %d->%d\n",(dlc_lists==DOA.data.hair_tbl)?'H':'F', from,to);

    struct doa_mdl *list = dlc_lists[info->char_idx];
    dlc_lists[info->char_idx]=malloc_copy(list,sizeof(struct doa_mdl),from,to);
    return 0;
}

static int set_dlc_entry(struct info *info, struct doa_mdl **dlc_lists,
             uint32_t idx, uint32_t *def)
{
    pr_dbg(".set dlc_def(%c,%d) ...\n",(dlc_lists==DOA.data.hair_tbl)?'H':'F', idx);

    struct doa_mdl* list = dlc_lists[info->char_idx];
    memcpy(list + idx, def, sizeof(struct doa_mdl));
    return 0;
}

static int set_alt_list_len(struct info *info, uint32_t new_len)
{
    struct cos_entry *cos = get_cos(info->char_idx, info->cos_idx);
    struct char_csopts_entry *opts = &DOA.data.opts_tbl[info->char_idx];
    uint8_t old_len = opts->num_alts[info->cos_idx];

    /* old_len is 1 even if there is no alt_tex used (alt_tbl==NULL)
     * which screws up our calculation */
    if(cos->alt_tbl==NULL) old_len=0;

    pr_dbg(" .num_alt(%s,cs:%d): %d->%d\n",DOA_CHAR_TBL[info->char_idx],
                                           info->cos_idx,old_len,new_len);

    if(new_len >0xff) {
       pr_log("*error* .nalt:  > 255: %d\n",new_len);
        return -1;
    }
    if( new_len > old_len) {
        cos->alt_tbl = malloc_copy(cos->alt_tbl, sizeof(struct alt_entry),
                                   old_len, new_len);
    }

    if(new_len==0) {
        cos->alt_tbl=NULL; // won't work otherwise
        opts->num_alts[info->cos_idx] = 1;
    } else {
        opts->num_alts[info->cos_idx] = (uint8_t)new_len;
    }
    return 0;
}

static int set_face_list_len(struct info *info, uint32_t num_faces)
{
    struct cos_entry *cos = get_cos(info->char_idx, info->cos_idx);
    struct char_csopts_entry *opts = &DOA.data.opts_tbl[info->char_idx];
    uint8_t old_len = opts->num_faces[info->cos_idx];

    /* old_len is 1 even if there is no alt_tex used (face_tbl==NULL)
     * which screws up our calculation */
    if(cos->face_tbl==NULL) old_len = 0;

    pr_dbg(" .num_faces(%s,cs:%d): %d->%d\n",DOA_CHAR_TBL[info->char_idx],
                                             info->cos_idx, old_len, num_faces);

    if(num_faces >0xff) {
        pr_log(".nfaces:  > 255: %d\n",num_faces);
        return -5;
    }
    uint8_t new_len = (uint8_t)num_faces;

    if( new_len > old_len) {
        cos->face_tbl = malloc_copy(cos->face_tbl, sizeof(struct face_entry),
                                    old_len, new_len);
    }
    if(new_len == 0) {
        cos->face_tbl = NULL;
        opts->num_faces[info->cos_idx] = 1;
    } else {
        opts->num_faces[info->cos_idx] = new_len;
    }
    return 0;
}

// face_def -> struct face_entry == 6*uint32
static int set_face(struct info *info, uint32_t face_idx, uint32_t *face_def)
{
    pr_dbg(" .set face_def(%d) ...\n",face_idx);
    struct cos_entry *cos = get_cos(info->char_idx, info->cos_idx);
    struct face_entry *face = cos->face_tbl + face_idx;

    memcpy(face, face_def, sizeof(*face));
    return 0;
}

// alt_def -> struct alt_entry == 2*uint32
static int set_alt(struct info *info, uint32_t alt_idx, uint32_t *alt_def)
{
    pr_dbg(" .set alt_def(%d) ...\n", alt_idx);
    struct cos_entry *cos = get_cos(info->char_idx, info->cos_idx);
    struct alt_entry *alt = cos->alt_tbl + alt_idx;

    memcpy(alt, alt_def, sizeof(uint32_t)*2);
    return 0;
}

static int swap_asset(struct info *info, uint32_t idx, uint32_t type,
              uint32_t flags, uint32_t size, const char *path)
{
    struct asset_entry *asset = DOA.data.asset_tbl + idx;
    pr_dbg("[swap] idx: %lx flags:%lx size:%lx path:%s\n",idx,flags,size,path);

    // setting lnk name to "NO_LINK" will make the game load the
    // external file ( without .lnk/.bin package )
    asset->name = strdup(path);
    asset->lnk = NOLINK;
    if(type != 0)
        asset->type = type;
    DOA.data.flag_tbl[idx].flags = flags;

    if( flags & FLAG_ZIP) // write unpacked size to crc_table
        DOA.data.crc_tbl[ CRC_TBL_IDX(flags) ].size = size;
    else
        DOA.data.flag_tbl[idx].size = size;
    return 0;
}

static int conv_ul(unsigned int num, char **src, uint32_t *dst)
{
    for(unsigned int i=0;i<num;i++) {
        errno=0;
        dst[i] = strtoul(src[i], NULL,0);
        if(errno!=0) return 1+i;
    }
    return 0;
}

#define MAX_IARGS 32

enum CFGCMD { VER, SWAP, SET_CHAR,SET_COS,CS_DEF,ENABLE_GLASSES,NUM_FACES,FACE_DEF,
              NUM_ALT, ALT_DEF, NUM_DLC_HAIR, NUM_DLC_FACE, DLC_HAIR, DLC_FACE };
static const struct cmd_desc {
    const char *cmd;
    unsigned int narg, int_arg;
} CMD_TBL[] = { {"ver",1,0}, {"swap",5, 4}, {"char", 1, 0}, {"cos",1,1}, {"cs_def",5,5},{"glasses",1,1},
         {"nfaces",1, 1}, {"face",7,7}, {"nalt",1,1}, {"alt",3,3}, {"ndlc_hair",2,2},
         {"ndlc_face",2,2},{"dlc_hair",3,3},{"dlc_face",3,3} };


static int conv_args(unsigned int lnum, unsigned int narg, char* args[], uint32_t *iargs)
{
    const struct cmd_desc *desc = &CMD_TBL[0];
    int ret;
    for(int i=0;i<ARRAY_SIZE(CMD_TBL);i++,desc++) {
        if(strcmp(args[0]+1, desc->cmd)) continue;
        if( (narg-1)!= desc->narg) {
            pr_log("*errror* mod.config:%d cmd:%s needs %d arguments\n",
                        lnum,desc->cmd,desc->narg);
            return -1;
        }
        if( (desc->int_arg)>MAX_IARGS) {
            pr_log("*error* mod.config:%d cmd:%s oops!this shouldn't happen\n",
                     lnum,desc->cmd);
            return -1;
        }
        ret = conv_ul(desc->int_arg,args+1, iargs);
        if(ret) {
            pr_log("*error* mod.config:%d cmd: %s failed to convert param(%d):%s\n",
                    lnum, desc->cmd, ret-1, args+ret);
            return -1;
        }
        return i;
    }
    pr_log("*warn* mod.config:%d unknown cmd: %s\n",args[0]+1);
    return -1;
}


static int apply_mod(unsigned int lnum, struct info *info, unsigned int narg, char* args[])
{
    uint32_t iargs[MAX_IARGS];
    if(info->fmt_ver == 1) {
        if( args[0][0]!='.' ) return 0; // ignore silently

        enum CFGCMD cmd = conv_args(lnum,narg, args, iargs);
        switch(cmd) {
        case VER:
            if( check_compat(info, args[1]) == 0)  return 0; // we're good
            else return -2; // abort, stop parsing
        case SWAP:
            return swap_asset(info, iargs[0], iargs[1], iargs[2], iargs[3],args[5]);
        case SET_CHAR:
            for(int i=0; i < MAX_CHARS; i++) {
                if(DOA_CHAR_TBL[i]==NULL) continue;
                if(strcmp(DOA_CHAR_TBL[i],args[1])==0) {
                    pr_dbg("=char: %s/%d\n",args[1],i);
                    info->char_idx=i;
                    return 0;
                }
            }
            pr_log("*error* mod.config:%d unknown character: %s",lnum,args[1]);
            return 0;
        case SET_COS:
            info->cos_idx=iargs[0];
            pr_dbg("=cos: %d\n",info->cos_idx);
            return 0;
        case CS_DEF: return set_cos_def(info, iargs);
        case ENABLE_GLASSES: return enable_glasses(info, iargs[0]);
        case NUM_FACES: return set_face_list_len(info, iargs[0]);
        case FACE_DEF: return set_face(info, iargs[0], iargs+1);
        case NUM_ALT: return set_alt_list_len(info, iargs[0]);
        case ALT_DEF: return set_alt(info, iargs[0], iargs+1);
        case NUM_DLC_HAIR:
            return set_dlc_list_len(info, DOA.data.hair_tbl, iargs[0],iargs[1]);
        case NUM_DLC_FACE:
            return set_dlc_list_len(info, DOA.data.face_tbl, iargs[0],iargs[1]);
        case DLC_HAIR:
            return set_dlc_entry(info, DOA.data.hair_tbl,iargs[0], iargs+1);
        case DLC_FACE:
            return set_dlc_entry(info, DOA.data.face_tbl,iargs[0], iargs+1);
        default: return 0; // continue with next line
        }
    } else if(info->fmt_ver==0) { // old fmt

        if(strcmp(args[0], ".fmt")==0) {
            if(narg<2) return -1;
            info->fmt_ver = atoi(args[1]);
            pr_log("mod fmt: %d\n",info->fmt_ver);
            return 0;
        }
        // old style fmt: <int idx>:<int flags>:<int size>:<char* path>
        if(narg==4) {
            uint32_t int_args[3];
            if(conv_ul(ARRAY_SIZE(int_args), args, int_args)==1) return -1;
            return swap_asset( info, int_args[0],0, int_args[1],int_args[2], args[3]);
        }
    }
    pr_log("*error* mod.config: unknown format\n");
    return -2; //abort,unknown fmt
}

static size_t get_file_size(const char *path)
{
    struct stat st={0};
    stat(path, &st);
    return st.st_size;
}

int load_config(struct info *info, const char *path)
{
    pr_log("loading mod.config: %s\n",path);
    FILE *f = fopen(path, "rb");
    if( f == 0 ) {
        pr_log("no mod.config exists\n");
        return 0;
    }
    size_t size = get_file_size(path);
    char *buf = (void*)malloc(size);
    size_t nb = fread(buf, 1, size, f);
    fclose(f);

    char *val = buf;
    int i = 0;
    int n=1;
    char* vals[20]={0};
    info->char_idx=-1;
    info->cos_idx=-1;
    for(char *p = buf; p < buf+nb; p++) {
        if( (*p == ':') || (*p == '\n') || (*p=='\r') ) {
            int endl = ( (*p=='\n')||(p==buf+nb-1))?1:0;
            *p ='\0';
            if(*val && (i < 20) )
                vals[i++]=val;
            if(endl & (i>0) ) {
                int res = apply_mod(n, info, i, vals);
                if(res==-1) pr_log("mod.config:%d error when applying cmd\n", n);
                if(res==-2) return -1;
                i=0;
                n++;
            }
           val = p+1;
        }
    }
    free(buf);
    return 0;
}

